#ifndef __qdrp_h__
#define __qdrp_h__

#include <qdrp/qdrp_global.h>
#include <qdrp/qdrp_imports.h>
#include <qdrp/qdrp_node.h>
#include <qdrp/qdrp_parent_selection_list.h>
#include <qdrp/qdrp_packet.h>
#include <qdrp/qdrp_address_map.h>

/* DECLARACIONES POR ADELANTADO DE LAS CLASES */
class QDRP;
class qdrp_state;
class qdrp_state__off;	
class qdrp_state__booting;
class qdrp_state__offline;
class qdrp_state__connecting;
class qdrp_state__online;
class qdrp_state__disconnecting;





// **************************************************************************************************************************************************
// TEMPORIZADORES
// **************************************************************************************************************************************************

// Temporizador para unirse a una red
class qdrp_timer__join_network: public TimerHandler
{
protected:
	QDRP* agent_;

	virtual void expire(Event* e);
public:
	qdrp_timer__join_network(QDRP* agent) : TimerHandler(){ agent_=agent; }
};


// Temporizador periódico para la emisión de SHOW_OFFs 
class qdrp_timer__heartbeat : public TimerHandler
{
protected:
	QDRP* agent_;

	virtual void expire(Event* e);
public:
	qdrp_timer__heartbeat(QDRP* agent) : TimerHandler(){ agent_=agent; }
};


// Temporizador para la conexión con el nodo padre
class qdrp_timer__parent_node_connection : public TimerHandler
{
protected:
	QDRP* agent_;
	u_int8_t cycle_;

	virtual void expire(Event* e);
public:
	qdrp_timer__parent_node_connection(QDRP* agent) : TimerHandler(){ agent_=agent; cycle_=0; }
	inline u_int8_t& cycle(){ return cycle_; }
};


// Temporizador para la conexión con un nodo hijo
class qdrp_timer__child_node_connection : public TimerHandler
{
protected:
	QDRP* agent_;
	u_int8_t child_id_;
	u_int8_t cycle_;

	virtual void expire(Event* e);
public:
	qdrp_timer__child_node_connection(QDRP* agent, u_int8_t child_id) : TimerHandler()
	{
		agent_=agent;
		child_id_=child_id;
		cycle_=0;
	}
	inline u_int8_t& cycle(){ return cycle_; }
};


//Temporizador para la emision de sendme show_off's

class qdrp_timer__sendme_show_off : public TimerHandler
{
protected:
	QDRP* agent_;

	virtual void expire(Event* e);
public:
	qdrp_timer__sendme_show_off(QDRP* agent) : TimerHandler(){ agent_=agent; }
};


//Temporizador para eligir una nueva/mejor routa hacia el sumidero

class qdrp_timer__new_route : public TimerHandler
{
protected:
	QDRP* agent_;

	virtual void expire(Event* e);
public:
	qdrp_timer__new_route(QDRP* agent) : TimerHandler(){ agent_=agent; }
};


// **************************************************************************************************************************************************
// FIN DE: TEMPORIZADORES
// **************************************************************************************************************************************************





// **************************************************************************************************************************************************
// ESTADO DE UN NODO
// **************************************************************************************************************************************************

class qdrp_state
{
protected:
	QDRP* agent_;

public: 
	//==========================================================================================================================//
	// Tipos de estado	 											     	    //
	//==========================================================================================================================//
	static const u_int8_t NUM_STATES = 6;
	
	static const u_int8_t OFF = 0;
	static const u_int8_t BOOTING = 1;
	static const u_int8_t OFFLINE = 2;
	static const u_int8_t CONNECTING = 3;
	static const u_int8_t ONLINE = 4;
	static const u_int8_t DISCONNECTING = 5;

	//==========================================================================================================================//
	// Constructor/Destructor 											     	    //
	//==========================================================================================================================//
	static qdrp_state* create_state(QDRP* system, u_int8_t state_type);
	virtual ~qdrp_state();

	//==========================================================================================================================//
	// Inicialización de estado											     	    //
	//==========================================================================================================================//
	virtual void initialize() = 0;

	//==========================================================================================================================//
	// Temporizadores												     	    //
	//==========================================================================================================================//
	virtual void On_timer__join_network() = 0;
	virtual void On_timer__heartbeat() = 0;
	virtual void On_timer__connection_with_parent_node() = 0;
	virtual void On_timer__connection_with_child_node(u_int8_t child_id) = 0;
	virtual void On_timer__sendme_show_off() = 0;	
	virtual void On_timer__new_route() = 0;

	//==========================================================================================================================//
	// Comandos													     	    //
	//==========================================================================================================================//
	virtual void On_command__boot() = 0;
	virtual void On_command__boot_sink(u_int8_t network_id) = 0;
	virtual void On_command__switch_off() = 0;	

	//==========================================================================================================================//
	// Recepción de paquetes											     	    //
	//==========================================================================================================================//
	virtual	void On_reception_of_data_packet(Packet* p) = 0;

	virtual void On_reception_of_qdrp_packet__show_off(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__connect(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__connection_accepted(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__connection_rejected(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__connection_renew(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__disconnect(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__disconnection_ok(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__route_crash(Packet* p) = 0;
	virtual void On_reception_of_qdrp_packet__sendme_show_off(Packet* p) = 0;	
//	virtual void On_reception_of_qdrp_packet__new_route(Packet* p) = 0;
};

// **************************************************************************************************************************************************
// FIN DE: ESTADO DE UN NODO
// **************************************************************************************************************************************************



// =============================================================================================================================================== //
// ESTADO = OFF
// =============================================================================================================================================== //

class qdrp_state__off : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__off(QDRP* agent);
	~qdrp_state__off();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off();
	void On_timer__new_route();
	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);
	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);
//	void On_reception_of_qdrp_packet__new_route(Packet* p);
};

// =============================================================================================================================================== //
// FIN DE: ESTADO = OFF
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = ARRANCANDO
// =============================================================================================================================================== //

class qdrp_state__booting : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__booting(QDRP* agent);
	~qdrp_state__booting();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off();
	void On_timer__new_route();

	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);
	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);	
//	void On_reception_of_qdrp_packet__new_route(Packet* p);

};

// =============================================================================================================================================== //
// FIN DE: ESTADO = ARRANCANDO
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = DESCONECTADO
// =============================================================================================================================================== //

class qdrp_state__offline : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__offline(QDRP* agent);
	~qdrp_state__offline();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off(); 
	void On_timer__new_route();

	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);
	
	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);	
//	void On_reception_of_qdrp_packet__new_route(Packet* p);
};
// =============================================================================================================================================== //
// FIN DE: ESTADO = DESCONECTADO
// =============================================================================================================================================== //


// =============================================================================================================================================== //
// ESTADO = CONECTANDO
// =============================================================================================================================================== //

class qdrp_state__connecting : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__connecting(QDRP* agent);
	~qdrp_state__connecting();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off();
	void On_timer__new_route();

	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);

	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);
//	void On_reception_of_qdrp_packet__new_route(Packet* p);
private:
	// ************************************************************************************************************************ //
	// Funciones internas												     	    //
	// ************************************************************************************************************************ //	
	void On_connection_attempt_failure();
	void On_connection_attempt_success();
};

// =============================================================================================================================================== //
// FIN DE: ESTADO = CONECTANDO
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = CONECTADO
// =============================================================================================================================================== //

class qdrp_state__online : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__online(QDRP* agent);
	~qdrp_state__online();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off();
	void On_timer__new_route();

	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);
	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);
//	void On_reception_of_qdrp_packet__new_route(Packet* p);
	
private:
	// ************************************************************************************************************************ //
	// Funciones internas												     	    //
	// ************************************************************************************************************************ //	
	void On_route_lost(bool timed_out);
};

// =============================================================================================================================================== //
// FIN DE: ESTADO = CONECTADO
// =============================================================================================================================================== //



// =============================================================================================================================================== //
// ESTADO = DESCONECTANDO
// =============================================================================================================================================== //

class qdrp_state__disconnecting : public qdrp_state
{
public:
	// ************************************************************************************************************************ //
	// Constructor/Destructor 											     	    //
	// ************************************************************************************************************************ //
	qdrp_state__disconnecting(QDRP* agent);
	~qdrp_state__disconnecting();

	// ************************************************************************************************************************ //
	// Inicialización de estado											     	    //
	// ************************************************************************************************************************ //
	void initialize();

	// ************************************************************************************************************************ //
	// Temporizadores												     	    //
	// ************************************************************************************************************************ //
	void On_timer__join_network();
	void On_timer__heartbeat();
	void On_timer__connection_with_parent_node();
	void On_timer__connection_with_child_node(u_int8_t child_id);
	void On_timer__sendme_show_off();
	void On_timer__new_route();

	// ************************************************************************************************************************ //
	// Comandos		 											     	    //
	// ************************************************************************************************************************ //
	void On_command__boot();
	void On_command__boot_sink(u_int8_t network_id);
	void On_command__switch_off();

	// ************************************************************************************************************************ //
	// Recepción de paquetes											     	    //
	// ************************************************************************************************************************ //	
	void On_reception_of_data_packet(Packet* p);
	void On_reception_of_qdrp_packet__show_off(Packet* p);
	void On_reception_of_qdrp_packet__connect(Packet* p);
	void On_reception_of_qdrp_packet__connection_accepted(Packet* p);
	void On_reception_of_qdrp_packet__connection_rejected(Packet* p);
	void On_reception_of_qdrp_packet__connection_renew(Packet* p);
	void On_reception_of_qdrp_packet__disconnect(Packet* p);
	void On_reception_of_qdrp_packet__disconnection_ok(Packet* p);
	void On_reception_of_qdrp_packet__route_crash(Packet* p);
	void On_reception_of_qdrp_packet__sendme_show_off(Packet* p);
//	void On_reception_of_qdrp_packet__new_route(Packet* p);
};

// =============================================================================================================================================== //
// FIN DE: ESTADO = DESCONECTANDO
// =============================================================================================================================================== //










// *********************************************************************************************************************************************** //
// AGENTE QDRP																	   //
// *********************************************************************************************************************************************** //
class QDRP : public Agent
{
	// ======================================================================================================================== //
	// Clases "amigas": temporizadores											    //
	// ======================================================================================================================== //
	friend class qdrp_timer__join_network;
	friend class qdrp_timer__heartbeat;
	friend class qdrp_timer__parent_node_connection;
	friend class qdrp_timer__child_node_connection;
	friend class qdrp_timer__sendme_show_off;
	friend class qdrp_timer__new_route;

	// ======================================================================================================================== //
	// Clases "amigas": estados												    //
	// ======================================================================================================================== //
	friend class qdrp_state;
	friend class qdrp_state__off;
	friend class qdrp_state__booting;
	friend class qdrp_state__offline;
	friend class qdrp_state__connecting;
	friend class qdrp_state__online;
	friend class qdrp_state__disconnecting;
	
	// ======================================================================================================================== //
	// Máquina de estados													    //
	// ======================================================================================================================== //
	qdrp_state* state_machine_[qdrp_state::NUM_STATES];		// vector de estados
	u_int8_t current_state_id_;					// estado actual
	
	// ======================================================================================================================== //
	// Situación del nodo en el árbol											    //
	// ======================================================================================================================== //
	qdrp_node*		this_node_;				// información referente a este nodo
	nsaddr_t		parent_node_address_;			// dirección IP del nodo padre
	qdrp_address_map*	child_node_address_map_;		// mapa con las direcciones IP de los nodos hijos

	// ======================================================================================================================== //
	// Lista de nodos descubiertos (utilizado en la fase de conexión) 							    //
	// ======================================================================================================================== //
	qdrp_parent_selection_list* 	discovered_node_list_;
	
	// ======================================================================================================================== //
	// Configuración del nodo												    //
	// ======================================================================================================================== //
	bool node_is_sink_;						// Indica si el nodo es un sumidero (1) o no (0)
	u_int8_t network_id_;						// ID de red (define una subred gobernada por un sumidero)

protected:
	// ======================================================================================================================== //
	// Temporizadores													    //
	// ======================================================================================================================== //
	qdrp_timer__join_network*			join_network_timer_;
	qdrp_timer__heartbeat*				heartbeat_timer_;
	qdrp_timer__parent_node_connection*		parent_node_connection_timer_;
	qdrp_timer__child_node_connection*		child_node_connection_timer_[QDRP_global::SINK__NUM_CHILD_SLOTS];
	qdrp_timer__sendme_show_off*			sendme_show_off_;
	qdrp_timer__new_route*				new_route_;
	// ======================================================================================================================== //
	// Objetos especiales													    //
	// ======================================================================================================================== //
	PortClassifier*		demultiplexer_;		// objeto utilizado pasar paquetes a otros agentes
	Trace*			log_target_;		//    "       "     para realizar una traza

	// ======================================================================================================================== //
	// Acceso a: Temporizadores												    //
	// ======================================================================================================================== //
	inline qdrp_timer__join_network*&		join_network_timer(){ return join_network_timer_; }
	inline qdrp_timer__heartbeat*& 			heartbeat_timer(){ return heartbeat_timer_; }
	inline qdrp_timer__parent_node_connection*& 	parent_node_connection_timer(){ return parent_node_connection_timer_; }
	inline qdrp_timer__child_node_connection* 	child_node_connection_timer(u_int8_t child_id)
	{
		return (child_id < QDRP_global::SINK__NUM_CHILD_SLOTS)? child_node_connection_timer_[child_id] : NULL;
	}

	inline qdrp_timer__sendme_show_off*&		sendme_show_off_timer(){ return sendme_show_off_; }
	
	inline qdrp_timer__new_route*&			new_route_timer(){ return new_route_; }
	// ======================================================================================================================== //
	// Acceso a: Máquina de estados												    //
	// ======================================================================================================================== //
	inline qdrp_state* current_state(){ return state_machine_[current_state_id_]; }
	inline u_int8_t& current_state_id(){ return current_state_id_; }
	void change_state(u_int8_t next_state_id);

	// ======================================================================================================================== //
	// Acceso a: Situación del nodo en el árbol										    //
	// ======================================================================================================================== //
	inline qdrp_node*& 		this_node(){ return this_node_; }
	inline nsaddr_t& 		parent_node_address(){ return parent_node_address_; }
	inline qdrp_address_map*&	child_node_address_map(){ return child_node_address_map_; }

	// ======================================================================================================================== //
	// Acceso a: Lista de nodos descubiertos										    //
	// ======================================================================================================================== //
	inline qdrp_parent_selection_list*& 	discovered_node_list(){ return discovered_node_list_; }

	// ======================================================================================================================== //
	// Acceso a: Configuración de simulación										    //
	// ======================================================================================================================== //
	inline bool& node_is_sink(){ return node_is_sink_; }
	inline u_int8_t& network_id(){ return network_id_; }

	// ======================================================================================================================== //
	// TXon y RXon de paquetes												    //
	// ======================================================================================================================== //
	Packet* create_qdrp_packet(nsaddr_t destination_address);	// Crea una plantilla de paquete QDRP vacío
	void On_reception_of_qdrp_packet(Packet* p);			// Evento de recepción de un paquete QDRP
	void On_reception_of_data_packet(Packet* p);			//   "    "      "     "   "    "    de datos

	void send_data_packet(Packet* p);				// Envía un paquete de datos
	void send_qdrp_broadcast_packet(Packet* p);			// Envía un paquete QDRP de difusión
	void send_qdrp_unicast_packet(Packet* p);			// Envía un paquete QDRP
	void resend_packet(Packet* p);					// Retransmite un paquete

	void send_qdrp_packet__show_off();					// Crea y envía un paquete SHOW_OFF
	void send_qdrp_packet__connect(nsaddr_t destination);			//   "  "   "    "    "    CONNECT
	void send_qdrp_packet__connection_accepted(nsaddr_t destination);	//   "  "   "    "    "    CONNECTION_ACCEPTED
	void send_qdrp_packet__connection_rejected(nsaddr_t destination);	//   "  "   "    "    "    CONNECTION_REJECTED
	void send_qdrp_packet__connection_renew(nsaddr_t destination);		//   "  "   "    "    "    CONNECTION_RENEW
	void send_qdrp_packet__disconnect(nsaddr_t destination);	 	//   "  "   "    "    "    DISCONNECT
	void send_qdrp_packet__disconnection_ok(nsaddr_t destination);		//   "  "   "    "    "    DISCONNECTION_OK
	void send_qdrp_packet__route_crash();					//   "  "   "    "    "    ROUTE_CRASH
	void send_qdrp_packet__sendme_show_off();		//   "  "   "    "    Sendme_show_off

	// ======================================================================================================================== //
	// Operaciones														    //
	// ======================================================================================================================== //
	void print_routing_info(Trace* out);	// Muestra la información sobre el árbol de nodos de la que dispone este nodo
	void switch_off();			// Apagado forzoso del nodo		
	
public:
	// ======================================================================================================================== //
	// Constructor														    //
	// ======================================================================================================================== //
	QDRP(nsaddr_t);

	// ======================================================================================================================== //
	// Operaciones de la clase Agent											    //
	// ======================================================================================================================== //
	int command(int, const char*const*);
	void recv(Packet*, Handler*);

	// ======================================================================================================================== //
	// Callbacks MAC													    //
	// ======================================================================================================================== //
	void On_mac_failure(Packet*);
	static void qdrp_mac_failed_callback(Packet*, void*);
};		

// *********************************************************************************************************************************************** //
// FIN DE: AGENTE QDRP
// *********************************************************************************************************************************************** //

#endif

