#include <qdrp/qdrp_address_map.h>

// **************************************************************************************************************************************************
// MAPA DE DIRECCIONES
// **************************************************************************************************************************************************

//==================================================================================================================================//
// Constructor/Destructor 													    //
//==================================================================================================================================//
// Constructor //
qdrp_address_map::qdrp_address_map(u_int8_t max_size)
{
	max_size_ = max_size;
	clear();
}

// Destructor //
qdrp_address_map::~qdrp_address_map(){}


// Borrado //
void qdrp_address_map::clear()
{
	// vacía las tablas de búsqueda directa e inversa, y carga la lista de identificadores 
	direct_seek_table_.clear();
	reverse_seek_table_.clear();
	identifier_queue_.clear();
	for(u_int8_t i = 0; i < max_size_; i++){ identifier_queue_.push_back(i); }
}


//==================================================================================================================================//
// Funciones de consulta													    //
//==================================================================================================================================//
bool qdrp_address_map::is_empty(){ return direct_seek_table_.empty(); }

bool qdrp_address_map::is_full(){ return direct_seek_table_.size() == max_size_; }

u_int8_t qdrp_address_map::size(){ return direct_seek_table_.size(); }

u_int8_t qdrp_address_map::max_size(){ return max_size_; }

u_int8_t qdrp_address_map::get_identifier_by_address(nsaddr_t address, bool* found)
{
	std::map<nsaddr_t, u_int8_t>::iterator it;
	u_int8_t identifier=0xFF;

	it=direct_seek_table_.find(address);

	*found = (it != direct_seek_table_.end())? true : false;
	if (*found)
	{
		identifier = it->second;
	}
	return identifier;
}

nsaddr_t qdrp_address_map::get_address_by_identifier(u_int8_t identifier, bool* found)
{
	std::map<u_int8_t, nsaddr_t>::iterator it;
	nsaddr_t address=IP_BROADCAST;

	it=reverse_seek_table_.find(identifier);
	
	*found = (it != reverse_seek_table_.end())? true : false;
	if(*found)
	{
		address = it->second;
	}
	return address;
}


void qdrp_address_map::dump_on_trace(Trace* out)
{
	std::map<nsaddr_t, u_int8_t>::iterator it_di;
	for(it_di=direct_seek_table_.begin() ; it_di!=direct_seek_table_.end(); it_di++)
	{
		sprintf(out->pt_->buffer(), "   - CHILD NODE(id: %d): %d", it_di->second, it_di->first);
		out->pt_->dump();
	}
}


//==================================================================================================================================//
// Modificadores														    //
//==================================================================================================================================//
// Añade una nueva dirección al mapa de direcciones //
void qdrp_address_map::add_entry(nsaddr_t address)
{
	assert(!is_full());
	
	std::map<nsaddr_t, u_int8_t>::iterator it;
	it=direct_seek_table_.find(address);

	// se asegura de no introducir una entrada repetida
	if(it == direct_seek_table_.end())
	{
		// obtiene un identificador para utilizar en las tablas directa e inversa
		u_int8_t identifier=identifier_queue_.front();
		identifier_queue_.pop_front();
	
		// introduce la entrada en las tablas de búsqueda directa e inversa 
		direct_seek_table_.insert(pair<nsaddr_t, u_int8_t>(address, identifier));
		reverse_seek_table_.insert(pair<u_int8_t, nsaddr_t>(identifier, address));
	}
}


// Elimina una entrada del mapa de direcciones, dada su dirección IP //
void qdrp_address_map::delete_entry(nsaddr_t address)
{
	assert(!is_empty());

	std::map<nsaddr_t, u_int8_t>::iterator it;
	it=direct_seek_table_.find(address);

	// hay cosas más importantes que hacer que borrar entradas inexistentes...
	if(it != direct_seek_table_.end())
	{
		// obtiene el identificador asociado a la entrada
		u_int8_t identifier = it->second;
		
		// borra la entrada en las tablas de búsqueda directa e inversa
		direct_seek_table_.erase(address);
		reverse_seek_table_.erase(identifier);
		
		// por último, libera el identificador
		identifier_queue_.push_back(identifier);
	}
}


// Elimina una entrada del mapa de direcciones, dado su identificador //
void qdrp_address_map::delete_entry_by_id(u_int8_t identifier)
{
	assert(!is_empty());

	std::map<u_int8_t, nsaddr_t>::iterator it;
	it=reverse_seek_table_.find(identifier);

	// se asegura de que el identificador exista para poder borrar la entrada asociada a éste
	if(it != reverse_seek_table_.end())
	{
		// obtiene la dirección asociada al identificador
		nsaddr_t address = it->second;
		
		// borra la entrada en las tablas de búsqueda directa e inversa
		direct_seek_table_.erase(address);
		reverse_seek_table_.erase(identifier);
		
		// y libera el identificador
		identifier_queue_.push_back(identifier);
	}
}


// **************************************************************************************************************************************************
// FIN DE: MAPA DE DIRECCIONES
// **************************************************************************************************************************************************

